from __future__ import annotations
from typing import Any, Optional, Iterator

from kojo.item import Item


class TabularReader:
    """A generator to provide kojo.Items from tabular data"""

    iterator: Iterator[list[Any]]
    headers: list[str]
    index: int
    default_cell_value: Any

    def __init__(
        self,
        inner_iterator: Iterator[list[Any]],
        default_cell_value: Optional[Any] = None,
    ) -> None:
        self.iterator = inner_iterator
        self.default_cell_value = default_cell_value
        self.headers = [str(h) for h in next(self.iterator)]
        self.index = 0

    def __next__(self) -> Item:
        row = next(self.iterator)
        lr = len(row)
        lh = len(self.headers)
        if lr < lh:
            # fill missing rows with default value
            for _ in range(lr, lh):
                row.append(self.default_cell_value)
        item = Item({self.headers[i]: cell for (i, cell) in enumerate(row)})

        item.meta["index"] = self.index
        self.index += 1

        return item

    def __iter__(self) -> TabularReader:
        return self



from __future__ import annotations
import json
from typing import Any, Callable, Iterable, Iterator

from kojo.item import Item


def apply(iterable: Iterable[Item], *processes: Process) -> Iterator[Item]:
    for item in iterable:
        processed_item = apply_on_item(item, *processes)
        if processed_item is not None:
            yield processed_item


def apply_on_item(item: Item | None, *processes: Process) -> Item | None:
    for process in processes:
        for step in process.steps:
            if item is None:
                return None
            item = step(item)
    return item


FilterStepFunction = Callable[[Item], bool]

MapStepFunction = Callable[[Item], Item]


class MapStep:
    fn: MapStepFunction

    def __init__(self, fn: MapStepFunction):
        self.fn = fn

    def __call__(self, item: Item) -> Item:
        return self.fn(item)


class FilterStep:
    fn: FilterStepFunction

    def __init__(self, fn: FilterStepFunction):
        self.fn = fn

    def __call__(self, item: Item) -> Item | None:
        keep = self.fn(item)
        return item if keep else None


class Process:
    steps: list[FilterStep | MapStep]

    def __init__(self) -> None:
        self.steps = []

    def map(self, fn: MapStepFunction) -> Process:
        self.steps.append(MapStep(fn))
        return self

    def filter(self, fn: FilterStepFunction) -> Process:
        self.steps.append(FilterStep(fn))
        return self

    def __call__(self, iterable: Iterable[Item]) -> Iterator[Item]:
        return apply(iterable, self)

    def __iadd__(self, fn: MapStepFunction) -> Process:
        return self.map(fn)

    def __imul__(self, fn: FilterStepFunction) -> Process:
        return self.filter(fn)


class StepEncoder(json.JSONEncoder):
    def default(self, step: FilterStep | MapStep) -> dict[str, Any]:
        fn = step.fn

        obj = {
            "type": step.__class__.__name__,
            "module": fn.__module__,
            "name": fn.__class__.__name__,
            "description": fn.__doc__,
        }
        if obj["name"] == "function":
            obj["name"] = fn.__name__

        return obj


class ProcessEncoder(json.JSONEncoder):
    def default(self, process: Process) -> list[dict[str, Any]]:
        child_encoder = StepEncoder()
        return [child_encoder.default(step) for step in process.steps]



# flake8: noqa

from kojo.item import (
    Item,
    ItemLog,
    ItemLogEntry,
    ItemDecoder,
    ItemEncoder,
    ItemLogEncoder,
    ItemLogEntryEncoder,
)

from kojo.process import (
    apply,
    apply_on_item,
    FilterStep,
    FilterStepFunction,
    MapStep,
    MapStepFunction,
    Process,
    StepEncoder,
    ProcessEncoder,
)



from __future__ import annotations
from collections import UserDict, UserList
import copy
import json
import logging
from typing import Any, Iterable, Union

DictType = Union[dict[str, Any], UserDict[str, Any]]


def _flatten(item: DictType, prefix: str = "") -> DictType:
    flat_dict = {}
    for k, v in item.items():
        kk = k
        if prefix:
            kk = prefix + "." + k

        if type(v) is dict:
            for ksub, vsub in _flatten(v, kk).items():
                flat_dict[ksub] = vsub
        else:
            flat_dict[kk] = v
    return flat_dict


class ItemLogEntry:
    level: int  # logging.NOTSET .. logging.CRITICAL
    message: str
    details: dict[str, Any]

    def __init__(self, _level: int, _message: str, **details: Any):
        self.level = _level
        self.message = _message
        self.details = details

    def __repr__(self) -> str:
        d = "".join([", {}={}".format(k, repr(v)) for k, v in self.details.items()])
        return "ItemLogEntry(logging.{}, {}{})".format(
            logging.getLevelName(self.level), repr(self.message), d
        )

    def __contains__(self, key: str) -> bool:
        return key in self.details

    def __getitem__(self, key: str) -> Any:
        return self.details[key]


class ItemLog(UserList[ItemLogEntry]):
    @property
    def level(self) -> int:
        if len(self) == 0:
            return logging.NOTSET
        return max([entry.level for entry in self])

    def log(self, _level: int, _message: str, **details: Any) -> None:
        entry = ItemLogEntry(_level, _message, **details)
        self.append(entry)

    def debug(self, _message: str, **details: Any) -> None:
        self.log(logging.DEBUG, _message, **details)

    def info(self, _message: str, **details: Any) -> None:
        self.log(logging.INFO, _message, **details)

    def warning(self, _message: str, **details: Any) -> None:
        self.log(logging.WARNING, _message, **details)

    def error(self, _message: str, **details: Any) -> None:
        self.log(logging.ERROR, _message, **details)

    def critical(self, _message: str, **details: Any) -> None:
        self.log(logging.CRITICAL, _message, **details)

    def clone(self) -> ItemLog:
        return copy.deepcopy(self)

    def __repr__(self) -> str:
        return "ItemLog(" + ", ".join([repr(entry) for entry in self]) + ")"

    def __add__(self, other: Iterable[ItemLogEntry]) -> ItemLog:
        if type(other) is not ItemLog:
            msg = f"TypeError: unsupported operand type(s) for +: 'ItemLog' and '{type(other).__name__}'"
            raise TypeError(msg)

        return ItemLog(list(self) + list(other))


class Item(UserDict[str, Any]):
    """
    Item extends the dictionary class.

    Item behaves like a dict but provides additional functionality

    Each Item has an log list to report comments and incidents found while
    transforming, for example validation errors or import notes.

    Each Item has an empty dict of meta when being created. meta is
    used to track import information as the name of the file or the line
    number.
    """

    log: ItemLog
    meta: dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.log = ItemLog()
        self.meta = {}

    def clone(self) -> Item:
        item = copy.deepcopy(self)
        item.log = self.log.clone()
        item.meta = copy.copy(self.meta)
        return item

    def flatten(self) -> Item:
        item = Item(_flatten(self))
        item.log = self.log.clone()
        item.meta = copy.copy(self.meta)
        return item


class ItemLogEntryEncoder(json.JSONEncoder):
    def default(self, entry: ItemLogEntry) -> dict[str, Any]:
        return {
            "level": entry.level,
            "message": entry.message,
            "details": entry.details,
        }


class ItemLogEncoder(json.JSONEncoder):
    def default(self, item_log: ItemLog) -> list[dict[str, Any]]:
        encoder = ItemLogEntryEncoder()
        return [encoder.default(entry) for entry in item_log]


class ItemEncoder(json.JSONEncoder):
    def default(self, item: Item) -> dict[str, Any]:
        return {
            "data": dict(item),
            "log": ItemLogEncoder().default(item.log),
            "meta": item.meta,
        }


class ItemDecoder(json.JSONDecoder):
    def decode(self, s: str) -> Item:  # type: ignore
        raw = super().decode(s)
        item = Item(raw["data"])
        item.meta = raw["meta"]
        item.log = ItemLog()
        for entry in raw["log"]:
            item.log.log(int(entry["level"]), entry["message"], **entry["details"])
        return item
